'''Unittests for dataset class.'''

import unittest
import numpy as np
import pandas as pd
from pandas.api.types import is_string_dtype
from pandas.api.types import is_numeric_dtype

import ensembleset.feature_methods as fm
import tests.dummy_dataframe as test_data


class TestFeatureMethods(unittest.TestCase):
    '''Tests feature engineering method functions.'''

    def setUp(self):
        '''Dummy DataFrames for tests.'''

        self.dummy_df = test_data.DUMMY_DF


    def test_onehot_encoding(self):
        '''Tests string features onehot encoder.'''

        train_df, test_df=fm.onehot_encoding(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['strings'],
            {'sparse_output': False}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        for feature in list(train_df.columns):
            self.assertTrue(is_numeric_dtype(train_df[feature]))

        for feature in list(test_df.columns):
            self.assertTrue(is_numeric_dtype(test_df[feature]))


    def test_ordinal_encoding(self):
        '''Tests string feature ordinal encoder.'''

        train_df, test_df=fm.ordinal_encoding(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            ['strings'],
            {
                'handle_unknown': 'use_encoded_value',
                'unknown_value': np.nan  
            }
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))
        self.assertFalse(is_string_dtype(train_df['strings']))
        self.assertFalse(is_string_dtype(test_df['strings']))


    def test_poly_features(self):
        '''Tests polynomial feature transformer.'''

        train_df, test_df=fm.poly_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'degree': 2}

        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_spline_features(self):
        '''Tests spline features transformer.'''

        train_df, test_df=fm.spline_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'n_knots': 2}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_log_features(self):
        '''Tests log features transformer.'''

        train_df, test_df=fm.log_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'base': '2'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.log_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'base': 'e'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.log_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'base': '10'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_ratio_features(self):
        '''Tests ratio feature transformer.'''

        train_df, test_df=fm.ratio_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'div_zero_value': np.nan}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_exponential_features(self):
        '''Tests exponential features transformer.'''

        train_df, test_df=fm.exponential_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'base': 'e'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.exponential_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'base': '2'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_sum_features(self):
        '''Tests sum features transformer.'''

        train_df, test_df=fm.sum_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'n_addends': 2}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.sum_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'n_addends': 4}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_difference_features(self):
        '''Tests difference features transformer.'''

        train_df, test_df=fm.difference_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'n_subtrahends': 2}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))

        train_df, test_df=fm.difference_features(
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'n_subtrahends': 4}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))



'''Unittests for dataset class.'''

import unittest
import h5py
import numpy as np
import pandas as pd
import ensembleset.dataset as ds
import tests.dummy_dataframe as test_data

# pylint: disable=protected-access

class TestDataSetInit(unittest.TestCase):
    '''Tests for main data set generator class initialization.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = test_data.DUMMY_DF

        self.dataset = ds.DataSet(
            label='floats_pos',
            train_data=self.dummy_df.copy(),
            test_data=self.dummy_df.copy(),
            string_features=['strings']
        )


    def test_class_arguments(self):
        '''Tests assignments of class attributes from user arguments.'''

        self.assertTrue(isinstance(self.dataset.label, str))
        self.assertTrue(isinstance(self.dataset.train_data, pd.DataFrame))
        self.assertTrue(isinstance(self.dataset.test_data, pd.DataFrame))
        self.assertTrue(isinstance(self.dataset.string_features, list))
        self.assertEqual(self.dataset.string_features[0], 'strings')

        with self.assertRaises(TypeError):
            ds.DataSet(
                label=2, # Bad label
                train_data=self.dummy_df.copy(),
                test_data=self.dummy_df.copy(),
                string_features=['strings']
            )

        with self.assertRaises(TypeError):
            ds.DataSet(
                label='float_pos',
                train_data='Not a Pandas Dataframe', # Bad train data
                test_data=self.dummy_df.copy(),
                string_features=['strings']
            )

        with self.assertRaises(TypeError):
            ds.DataSet(
                label='float_pos',
                train_data=self.dummy_df.copy(),
                test_data='Not a Pandas Dataframe', # Bad test data
                string_features=['strings']
            )

        with self.assertRaises(TypeError):
            ds.DataSet(
                label='float_pos',
                train_data=self.dummy_df.copy(),
                test_data=self.dummy_df.copy(),
                string_features='Not a list of features' # Bad string features
            )


    def test_label_assignment(self):
        '''Tests assigning and saving labels.'''

        self.assertEqual(self.dataset.train_labels[-1], 7.0)
        self.assertEqual(self.dataset.test_labels[-1], 7.0)

        dataset=ds.DataSet(
            label='bad_label_feature',
            train_data=self.dummy_df,
            test_data=self.dummy_df,
            string_features=['strings']
        )

        self.assertTrue(np.isnan(dataset.train_labels[-1]))
        self.assertTrue(np.isnan(dataset.test_labels[-1]))


    def test_output_creation(self):
        '''Tests the creation of the HDF5 output sink.'''

        hdf = h5py.File('data/dataset.h5', 'r')

        self.assertTrue('train' in hdf)
        self.assertTrue('test' in hdf)
        self.assertEqual(hdf['test/labels'][-1], 7.0)
        self.assertEqual(hdf['test/labels'][-1], 7.0)

        hdf.close()

        _=ds.DataSet(
            label='bad_label_feature',
            train_data=self.dummy_df,
            test_data=self.dummy_df,
            string_features=['strings']
        )

        hdf = h5py.File('data/dataset.h5', 'r')

        self.assertTrue('train' in hdf)
        self.assertTrue('test' in hdf)
        self.assertTrue(np.isnan(hdf['test/labels'][-1]))
        self.assertTrue(np.isnan(hdf['test/labels'][-1]))

        hdf.close()


    def test_pipeline_options(self):
        '''Tests the creation of feature engineering pipeline options'''

        self.assertTrue(isinstance(self.dataset.string_encodings, dict))
        self.assertTrue(isinstance(self.dataset.numerical_methods, dict))


class TestDataPipelineGen(unittest.TestCase):
    '''Tests for data pipeline generator function.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = pd.DataFrame({
            'feature1': [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset = ds.DataSet(
            'feature2',
            self.dummy_df,
            test_data=self.dummy_df,
            string_features=['feature3']
        )


    def test_generate_data_pipeline(self):
        '''Tests the data pipeline generation function.'''

        pipeline=self.dataset._generate_data_pipeline(2)

        self.assertEqual(len(pipeline), 3)

        for operation, parameters in pipeline.items():
            self.assertTrue(isinstance(operation, str))
            self.assertTrue(isinstance(parameters, dict))


class TestFeatureSelection(unittest.TestCase):
    '''Tests for data pipeline generator function.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = pd.DataFrame({
            1: [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset = ds.DataSet(
            'feature2',
            self.dummy_df,
            test_data=self.dummy_df,
            string_features=['feature3']
        )


    def test_select_features(self):
        '''Tests feature selection function.'''

        features=self.dataset._select_features(3, self.dummy_df)

        self.assertEqual(len(features), 2)

        for feature in features:
            self.assertTrue(isinstance(feature, str))

        self.assertFalse('feature2' in features)


class TestDatasetGeneration(unittest.TestCase):
    '''Tests dataset generation.'''

    def setUp(self):
        '''Dummy DataFrames and datasets for tests.'''

        self.dummy_df = pd.DataFrame({
            1: [0,1],
            'feature2': [3,4],
            'feature3': ['a', 'b']
        })

        self.dataset = ds.DataSet(
            'feature2',
            self.dummy_df,
            test_data=self.dummy_df,
            string_features=['feature3']
        )

        self.dataset.make_datasets(2, 2, 1)


    def test_make_datasets(self):
        '''Tests generation of datasets.'''

        hdf = h5py.File('data/dataset.h5', 'a')

        training_datasets=hdf['train']
        self.assertEqual(len(training_datasets), 2)

        testing_datasets=hdf['test']
        self.assertEqual(len(testing_datasets), 2)



'''Dictionaries containing feature engineering methods for string in
numeric features and parameter argument options.'''

import numpy as np

# String feature encoding methods
STRING_ENCODINGS={
    'onehot_encoding': {'sparse_output': False},
    'ordinal_encoding': {
        'handle_unknown': 'use_encoded_value',
        'unknown_value': np.nan  
    }
}

# Numeric feature engineering methods
NUMERICAL_METHODS={
    'poly_features': {
        'degree': [2, 3],
        'interaction_only': [True, False],
    },
    'spline_features': {
        'n_knots': [5],
        'degree': [2, 3, 4],
        'knots': ['uniform', 'quantile'],
        'extrapolation': ['error', 'constant', 'linear', 'continue', 'periodic']
    },
    'log_features': {
        'base': ['2', 'e', '10']
    },
    'ratio_features': {
        'div_zero_value': [np.nan]
    },
    'exponential_features': {
        'base': ['2', 'e']
    },
    'sum_features': {
        'n_addends': [2,3,4]
    },
    'difference_features': {
        'n_subtrahends': [2,3,4]
    }
}


'''Generates variations of a dataset using a pool of feature engineering
techniques. Used for training ensemble models.'''

from pathlib import Path
from random import choice, shuffle

import h5py
import numpy as np
import pandas as pd

import ensembleset.feature_engineerings as engineerings
import ensembleset.feature_methods as fm


class DataSet:
    '''Dataset generator class.'''

    def __init__(
            self,
            label: str,
            train_data: pd.DataFrame,
            test_data: pd.DataFrame = None,
            string_features: list = None
        ):

        # Check user argument types
        type_check = self.check_argument_types(
            label,
            train_data,
            test_data,
            string_features
        )

        # If the type check passed, assign arguments to attributes
        if type_check is True:
            self.label = label
            self.train_data = train_data
            self.test_data = test_data
            self.string_features = string_features

        # Enforce string type on DataFrame columns
        self.train_data.columns = self.train_data.columns.astype(str)

        if self.test_data is not None:
            self.test_data.columns = self.test_data.columns.astype(str)

        # Retrieve and assign the training labels, set NAN if they don't exist
        # then remove them from the training data
        if self.label in self.train_data.columns:
            self.train_labels=np.array(self.train_data[label])
            self.train_data.drop(self.label, axis=1, inplace=True)

        else:
            self.train_labels=[np.nan] * len(self.train_data)

        # Retrieve and assign the testing labels, set NAN if they don't exist
        # then remove them from the training data
        if self.label in self.test_data.columns:
            self.test_labels=np.array(self.test_data[label])
            self.test_data.drop(self.label, axis=1, inplace=True)

        else:
            self.test_labels=[np.nan] * len(self.test_data)

        # Create the HDF5 output
        Path('data').mkdir(parents=True, exist_ok=True)

        # Create groups for training and testing datasets
        with h5py.File('data/dataset.h5', 'a') as hdf:

            _ = hdf.require_group('train')
            _ = hdf.require_group('test')

        # Add the training and testing labels
        with h5py.File('data/dataset.h5', 'w') as hdf:
            _ = hdf.create_dataset('train/labels', data=self.train_labels)
            _ = hdf.create_dataset('test/labels', data=self.test_labels)

        # Define the feature engineering pipeline methods
        self.string_encodings=engineerings.STRING_ENCODINGS
        self.numerical_methods=engineerings.NUMERICAL_METHODS


    def make_datasets(self, n_datasets:int, n_features:int, n_steps:int):
        '''Makes n datasets with different feature subsets and pipelines.'''

        hdf = h5py.File('data/dataset.h5', 'w')

        # Generate n datasets
        for n in range(n_datasets):

            print(f'\nGenerating dataset {n+1} of {n_datasets}')

            # Take a copy of the training and test data
            train_df = self.train_data.copy()
            test_df = self.test_data.copy()

            # Generate a data pipeline
            pipeline = self._generate_data_pipeline(n_steps)

            # Loop on and apply each method in the pipeline
            for method, arguments in pipeline.items():

                print(f' Applying {method}')
                func = getattr(fm, method)

                if method in self.string_encodings:
                    train_df, test_df = func(
                        train_df,
                        test_df,
                        self.string_features,
                        arguments
                    )

                else:
                    features = self._select_features(n_features, train_df)

                    train_df, test_df = func(
                        train_df,
                        test_df,
                        features,
                        arguments
                    )

            # Save the results to HDF5 output
            _ = hdf.create_dataset(f'train/{n}', data=np.array(train_df))
            _ = hdf.create_dataset(f'test/{n}', data=np.array(test_df))

        hdf.close()


    def _select_features(self, n_features:int, data_df:pd.DataFrame):
        '''Selects a random subset of features.'''

        features = data_df.columns.to_list()
        shuffle(features)
        features = features[:n_features]

        return features


    def _generate_data_pipeline(self, n_steps:int):
        '''Generates one random sequence of feature engineering operations. Starts with
        a string encoding method if we have string features.'''

        pipeline={}

        # Choose a string encoding method, if needed
        if self.string_features is not None:
            options = list(self.string_encodings.keys())
            selection = choice(options)
            pipeline[selection] = self.string_encodings[selection]

        # Construct a random sequence of numerical feature engineering methods
        methods = list(self.numerical_methods.keys())
        shuffle(methods)
        methods = methods[:n_steps]

        for method in methods:

            pipeline[method] = {}
            parameters = self.numerical_methods[method]

            for parameter, values in parameters.items():

                value = choice(values)
                pipeline[method][parameter] = value

        return pipeline


    def check_argument_types(self,
            label: str,
            train_data: pd.DataFrame,
            test_data: pd.DataFrame = None,
            string_features: list = None
    ) -> bool:

        '''Checks user argument types, returns true or false for all passing.'''

        check_pass = False

        if isinstance(label, str):
            check_pass = True

        else:
            raise TypeError('Label is not a string.')

        if isinstance(train_data, pd.DataFrame):
            check_pass = True

        else:
            raise TypeError('Train data is not a Pandas DataFrame.')

        if isinstance(test_data, pd.DataFrame) or test_data is None:
            check_pass = True

        else:
            raise TypeError('Test data is not a Pandas DataFrame.')

        if isinstance(string_features, list) or string_features is None:
            check_pass = True

        else:
            raise TypeError('String features is not a list.')

        return check_pass



'''Collection of functions to run feature engineering operations.'''

from math import e
from itertools import permutations, combinations
from typing import Tuple

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, OrdinalEncoder, PolynomialFeatures, SplineTransformer


def onehot_encoding(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's one hot encoder.'''

    encoder=OneHotEncoder(**kwargs)

    encoded_data=encoder.fit_transform(train_df[features])
    encoded_df=pd.DataFrame(encoded_data, columns=encoder.get_feature_names_out())
    train_df.drop(features, axis=1, inplace=True)
    train_df=pd.concat([train_df, encoded_df], axis=1)

    if test_df is not None:
        encoded_data=encoder.transform(test_df[features])
        encoded_df=pd.DataFrame(encoded_data, columns=encoder.get_feature_names_out())
        test_df.drop(features, axis=1, inplace=True)
        test_df=pd.concat([test_df, encoded_df], axis=1)

    return train_df, test_df


def ordinal_encoding(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's label encoder.'''

    encoder=OrdinalEncoder(**kwargs)

    train_df[features]=encoder.fit_transform(train_df[features])

    if test_df is not None:
        test_df[features]=encoder.transform(test_df[features])

    return train_df, test_df


def poly_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's polynomial feature transformer..'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    transformer=PolynomialFeatures(**kwargs)
    imputer=SimpleImputer(strategy='mean')
    scaler=MinMaxScaler(feature_range=(0, 1))

    imputed_data=imputer.fit_transform(train_working_df[numeric_features])
    encoded_data=transformer.fit_transform(imputed_data)
    new_columns=transformer.get_feature_names_out()
    encoded_df=pd.DataFrame(encoded_data, columns=new_columns)
    encoded_df[new_columns]=scaler.fit_transform(encoded_df[new_columns])
    train_df.drop(features, axis=1, inplace=True)
    train_df=pd.concat([train_df, encoded_df], axis=1)

    if test_df is not None:

        imputed_data=imputer.transform(test_working_df[numeric_features])
        encoded_data=transformer.transform(imputed_data)
        new_columns=transformer.get_feature_names_out()
        encoded_df=pd.DataFrame(encoded_data, columns=new_columns)
        encoded_df[new_columns]=scaler.fit_transform(encoded_df[new_columns])
        test_df.drop(features, axis=1, inplace=True)
        test_df=pd.concat([test_df, encoded_df], axis=1)

    return train_df, test_df


def spline_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's polynomial feature transformer..'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    transformer=SplineTransformer(**kwargs)
    imputer=SimpleImputer(strategy='mean')
    scaler=MinMaxScaler(feature_range=(0, 1))

    imputed_data=imputer.fit_transform(train_working_df[numeric_features])
    encoded_data=transformer.fit_transform(imputed_data)
    new_columns=transformer.get_feature_names_out()
    encoded_df=pd.DataFrame(encoded_data, columns=new_columns)
    encoded_df[new_columns]=scaler.fit_transform(encoded_df[new_columns])
    train_df.drop(features, axis=1, inplace=True)
    train_df=pd.concat([train_df, encoded_df], axis=1)

    if test_df is not None:

        imputed_data=imputer.transform(test_working_df[numeric_features])
        encoded_data=transformer.transform(imputed_data)
        new_columns=transformer.get_feature_names_out()
        encoded_df=pd.DataFrame(encoded_data, columns=new_columns)
        encoded_df[new_columns]=scaler.fit_transform(encoded_df[new_columns])
        test_df.drop(features, axis=1, inplace=True)
        test_df=pd.concat([test_df, encoded_df], axis=1)

    return train_df, test_df


def log_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Takes log of feature, uses sklearn min-max scaler if needed
    to avoid undefined log errors.'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    imputer=SimpleImputer(strategy='mean')
    train_working_df[numeric_features] = imputer.fit_transform(train_working_df[numeric_features])
    test_working_df[numeric_features] = imputer.transform(test_working_df[numeric_features])


    for feature in numeric_features:
        if min(train_working_df[feature]) <= 0 or min(test_working_df[feature]) <= 0:

            scaler=MinMaxScaler(feature_range=(1, 100))

            train_working_df[feature]=scaler.fit_transform(train_working_df[feature].to_frame())
            test_working_df[feature]=scaler.transform(test_working_df[feature].to_frame())

        if kwargs['base'] == '2':
            train_df[f'{feature}_log2']=np.log2(train_working_df[feature])
            test_df[f'{feature}_log2']=np.log2(test_working_df[feature])

        if kwargs['base'] == 'e':
            train_df[f'{feature}_ln']=np.log(train_working_df[feature])
            test_df[f'{feature}_ln']=np.log(test_working_df[feature])

        if kwargs['base'] == '10':
            train_df[f'{feature}_log10']=np.log10(train_working_df[feature])
            test_df[f'{feature}_log10']=np.log10(test_working_df[feature])

    return train_df, test_df


def ratio_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds every possible ratio feature, replaces divide by zero errors
    with np.nan.'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    imputer=SimpleImputer(strategy='mean')
    train_working_df[numeric_features] = imputer.fit_transform(train_working_df[numeric_features])
    test_working_df[numeric_features] = imputer.transform(test_working_df[numeric_features])

    feature_pairs=permutations(numeric_features, 2)

    train_features={}
    test_features={}

    for feature_a, feature_b in feature_pairs:

        quotient = np.divide(
            np.array(train_working_df[feature_a]),
            np.array(train_working_df[feature_b]),
            out=np.array([kwargs['div_zero_value']]*len(train_working_df[feature_a])),
            where=np.array(train_working_df[feature_b]) != 0
        )

        train_features[f'{feature_a}_over_{feature_b}'] = quotient

        quotient = np.divide(
            np.array(test_working_df[feature_a]),
            np.array(test_working_df[feature_b]),
            out=np.array([kwargs['div_zero_value']]*len(test_working_df[feature_a])),
            where=np.array(test_working_df[feature_b]) != 0
        )

        test_features[f'{feature_a}_over_{feature_b}'] = quotient

    new_train_df=pd.DataFrame.from_dict(train_features)
    new_test_df=pd.DataFrame.from_dict(test_features)

    train_df=pd.concat([train_df, new_train_df], axis=1)
    test_df=pd.concat([test_df, new_test_df], axis=1)

    return train_df, test_df


def exponential_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds exponential features with base 2 or base e.'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    imputer=SimpleImputer(strategy='mean')
    train_working_df[numeric_features] = imputer.fit_transform(train_working_df[numeric_features])
    test_working_df[numeric_features] = imputer.transform(test_working_df[numeric_features])

    new_train_features={}
    new_test_features={}
    scaler=MinMaxScaler(feature_range=(0, 1))

    for feature in numeric_features:
        if kwargs['base'] == 'e':
            new_train_features[f'{feature}_exp_base_e'] = e**train_working_df[feature].astype(float)
            new_test_features[f'{feature}_exp_base_e'] = e**test_working_df[feature].astype(float)

        elif kwargs['base'] == '2':
            new_train_features[f'{feature}_exp_base_2'] = 2**train_working_df[feature].astype(float)
            new_test_features[f'{feature}_exp_base_2'] = 2**test_working_df[feature].astype(float)

    new_train_df=pd.DataFrame.from_dict(new_train_features)
    new_test_df=pd.DataFrame.from_dict(new_test_features)

    train_df=pd.concat([train_df, new_train_df], axis=1)
    test_df=pd.concat([test_df, new_test_df], axis=1)

    train_df[list(new_train_features.keys())]=scaler.fit_transform(
        train_df[list(new_train_features.keys())]
    )

    train_df[list(new_test_features.keys())]=scaler.fit_transform(
        train_df[list(new_test_features.keys())]
    )

    return train_df, test_df


def sum_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds sum features for variable number of addends.'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    imputer=SimpleImputer(strategy='mean')
    train_working_df[numeric_features] = imputer.fit_transform(train_working_df[numeric_features])
    test_working_df[numeric_features] = imputer.transform(test_working_df[numeric_features])

    if kwargs['n_addends'] > len(numeric_features):
        n_addends=len(numeric_features)

    else:
        n_addends=kwargs['n_addends']

    new_test_features={}
    new_train_features={}
    addend_sets=combinations(numeric_features, n_addends)

    for i, addend_set in enumerate(addend_sets):

        train_sum = [0]*len(train_working_df)
        test_sum = [0]*len(test_working_df)

        for addend in addend_set:

            train_sum += train_working_df[addend]
            test_sum += test_working_df[addend]

        new_train_features[f'sum_feature_{i}'] = train_sum
        new_test_features[f'sum_feature_{i}'] = test_sum

    new_train_df=pd.DataFrame.from_dict(new_train_features)
    new_test_df=pd.DataFrame.from_dict(new_test_features)

    train_df=pd.concat([train_df, new_train_df], axis=1)
    test_df=pd.concat([test_df, new_test_df], axis=1)

    return train_df, test_df


def difference_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds difference features for variable number of subtrahends.'''

    # Exclude string features if present
    numeric_features=[]
    for feature in features:
        if is_numeric_dtype(train_df[feature]) and is_numeric_dtype(test_df[feature]):
            numeric_features.append(feature)

    train_working_df=train_df.copy()
    test_working_df=test_df.copy()

    train_working_df[numeric_features]=train_working_df[numeric_features].astype(float).copy()
    test_working_df[numeric_features]=test_working_df[numeric_features].astype(float).copy()

    # Get rid of np.inf
    train_working_df[numeric_features]=train_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )
    test_working_df[numeric_features]=test_working_df[numeric_features].replace(
        [np.inf, -np.inf],
        np.nan
    )

    # Get rid of large values
    train_working_df[numeric_features] = train_working_df[numeric_features].mask(
        train_working_df[numeric_features] > 5.0*10**102
    )
    test_working_df[numeric_features] = test_working_df[numeric_features].mask(
        test_working_df[numeric_features] > 5.0*10**102
    )

    imputer=SimpleImputer(strategy='mean')
    train_working_df[numeric_features] = imputer.fit_transform(train_working_df[numeric_features])
    test_working_df[numeric_features] = imputer.transform(test_working_df[numeric_features])

    if kwargs['n_subtrahends'] > len(numeric_features):
        n_subtrahends=len(numeric_features)

    else:
        n_subtrahends=kwargs['n_subtrahends']

    new_test_features={}
    new_train_features={}
    subtrahend_sets=combinations(numeric_features, n_subtrahends)

    for i, subtrahend_set in enumerate(subtrahend_sets):

        train_difference = train_working_df[subtrahend_set[0]]
        test_difference = test_working_df[subtrahend_set[0]]

        for subtrahend in subtrahend_set[1:]:

            train_difference -= train_working_df[subtrahend]
            test_difference -= test_working_df[subtrahend]

        new_train_features[f'difference_feature_{i}'] = train_difference
        new_test_features[f'difference_feature_{i}'] = test_difference

    new_train_df=pd.DataFrame.from_dict(new_train_features)
    new_test_df=pd.DataFrame.from_dict(new_test_features)

    train_df=pd.concat([train_df, new_train_df], axis=1)
    test_df=pd.concat([test_df, new_test_df], axis=1)

    return train_df, test_df
